# -*- coding: utf-8 -*-
"""
https://github.com/RDFLib/pySHACL/issues/162
"""

import rdflib
from pyshacl import validate

shacl_file = """\
@prefix exShapes: <urn:exShapes#> .
@prefix exOnt: <urn:exOnt#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix dcat: <http://www.w3.org/ns/dcat#> .
@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix vcard: <http://www.w3.org/2006/vcard/ns#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix schema: <http://schema.org/> .

exShapes:Example-shape
	a sh:NodeShape ;
	sh:targetClass exOnt:Dataset ;
	sh:property [ a sh:PropertyShape ;
		sh:maxCount "1"^^xsd:integer ;
		sh:minCount "1"^^xsd:integer ;
		sh:path dcat:contactPoint ;
		sh:qualifiedMaxCount 1;
		sh:qualifiedMinCount 1;
		sh:qualifiedValueShape [ a sh:NodeShape ;
			sh:class vcard:Contact ;
			sh:property [ a sh:PropertyShape ;
				sh:maxCount 1;
				sh:minCount 1;
				sh:path vcard:hasEmail ;
			] ,
			[ a sh:PropertyShape ;
				sh:maxCount 1;
				sh:path vcard:fn ;
			] ;
		] ;
	] .
"""

ont_file1 = """\
@prefix exOnt: <urn:exOnt#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix dcat: <http://www.w3.org/ns/dcat#> .
@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix vcard: <http://www.w3.org/2006/vcard/ns#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

<urn:exOnt> a owl:Ontology .

exOnt:Dataset a owl:Class .

exOnt:contact-0
	a vcard:Contact ,
	 vcard:Individual ,
	 owl:NamedIndividual ;
	rdfs:comment "This named individual is provided as a stand-in default"@en ;
	vcard:fn "NOT ENCODED" ;
	vcard:hasEmail <mailto:null@example.org> .

exOnt:canBeNamed a owl:Class ;
    exOnt:mightHave exOnt:contact-0 .
"""

ont_file2 = """\
@prefix exOnt: <urn:exOnt#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix dcat: <http://www.w3.org/ns/dcat#> .
@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix vcard: <http://www.w3.org/2006/vcard/ns#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

<urn:exOnt> a owl:Ontology .

exOnt:Dataset a owl:Class .

exOnt:contact-0
	a vcard:Contact ,
	 vcard:Individual ;
	rdfs:comment "This named individual is provided as a stand-in default"@en ;
	vcard:fn "NOT ENCODED" ;
	vcard:hasEmail <mailto:null@example.org> .
"""
data_file = """\
@prefix ex: <urn:ex#> .
@prefix exOnt: <urn:exOnt#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix dcat: <http://www.w3.org/ns/dcat#> .
@prefix schema: <http://schema.org/> .

ex:test_dataset a exOnt:Dataset ;
    dcat:contactPoint exOnt:contact-0 .
"""


def test_170() -> None:
    data_g = rdflib.Graph()
    data_g.parse(data=data_file, format="turtle")
    ont_g = rdflib.Graph()
    ont_g.parse(data=ont_file1, format="turtle")
    shapes = rdflib.Graph()
    shapes.parse(data=shacl_file, format="turtle")

    conforms, report, message = validate(data_g, shacl_graph=shapes, ont_graph=ont_g, inference="rdfs", debug=False)
    assert conforms

    ont_g2 = rdflib.Graph()
    ont_g2.parse(data=ont_file2, format="turtle")
    conforms, report, message = validate(data_g, shacl_graph=shapes, ont_graph=ont_g2, inference="rdfs", debug=True)
    assert not conforms


shacl_file_sparql_prefixes = '''# baseURI: urn:exShapes
@prefix exShapes: <urn:exShapes#> .
@prefix exOnt: <urn:exOnt#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix dcat: <http://www.w3.org/ns/dcat#> .
@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix vcard: <http://www.w3.org/2006/vcard/ns#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix schema: <http://schema.org/> .

<urn:exShapes>  rdf:type owl:Ontology ;
    sh:declare [
      rdf:type sh:PrefixDeclaration ;
      sh:namespace "http://www.w3.org/2006/vcard/ns#"^^xsd:anyURI ;
      sh:prefix "vcard" ;
    ] ;
    sh:declare [
      rdf:type sh:PrefixDeclaration ;
      sh:namespace "http://www.w3.org/ns/dcat#"^^xsd:anyURI ;
      sh:prefix "dcat" ;
    ] ;
    sh:declare [
      rdf:type sh:PrefixDeclaration ;
      sh:namespace "urn:exOnt#"^^xsd:anyURI ;
      sh:prefix "exOnt" ;
    ] ;
    .

exShapes:Example-shape
	a sh:NodeShape ;
	sh:targetClass exOnt:Dataset ;
  sh:sparql [
      rdfs:comment "Test using prefixes in sparql" ;
      sh:prefixes <urn:exShapes> ;
      sh:select """SELECT $this
WHERE {
	$this dcat:contactPoint ?c .
	FILTER NOT EXISTS {
		?c vcard:fn ?fn .
    }
}""" ;
    ] ;
    .
'''


def test_170_2() -> None:
    data_g = rdflib.Graph()
    data_g.parse(data=data_file, format="turtle")
    ont_g = rdflib.Graph()
    ont_g.parse(data=ont_file1, format="turtle")
    shapes = rdflib.Graph()
    shapes.parse(data=shacl_file_sparql_prefixes, format="turtle")

    conforms, report, message = validate(data_g, shacl_graph=shapes, ont_graph=ont_g, inference="rdfs", debug=True)
    assert conforms
    ont_2 = rdflib.Graph()
    ont_2.parse(data=ont_file2, format="turtle")
    conforms, report, message = validate(data_g, shacl_graph=shapes, ont_graph=ont_2, inference="rdfs", debug=True)
    assert not conforms


shacl_file_no_sparql_prefixes = '''# baseURI: urn:exShapes
@prefix exShapes: <urn:exShapes#> .
@prefix exOnt: <urn:exOnt#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix dcat: <http://www.w3.org/ns/dcat#> .
@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix vcard: <http://www.w3.org/2006/vcard/ns#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix schema: <http://schema.org/> .

<urn:exShapes>  rdf:type owl:Ontology .

exShapes:Example-shape
	a sh:NodeShape ;
	sh:targetClass exOnt:Dataset ;
  sh:sparql [
      rdfs:comment "Test using prefixes in sparql" ;
      sh:select """SELECT $this
WHERE {
	$this dcat:contactPoint ?c .
	FILTER NOT EXISTS {
		?c vcard:fn ?fn .
    }
}""" ;
    ] ;
    .
'''


def test_170_3() -> None:
    # The vcard prefix does not exist in the DataFile, only in the Ontology file.
    # It is defacto behaviour that PySHACL now copies any unknown prefixes from the ontology file into the data file.
    data_g = rdflib.Graph()
    data_g.parse(data=data_file, format="turtle")
    ont_g = rdflib.Graph()
    ont_g.parse(data=ont_file1, format="turtle")
    shapes = rdflib.Graph()
    shapes.parse(data=shacl_file_no_sparql_prefixes, format="turtle")
    conforms, report, message = validate(data_g, shacl_graph=shapes, ont_graph=ont_g, inference="rdfs", debug=True)
    assert conforms
    ont_2 = rdflib.Graph()
    ont_2.parse(data=ont_file2, format="turtle")
    conforms, report, message = validate(data_g, shacl_graph=shapes, ont_graph=ont_2, inference="rdfs", debug=True)
    assert not conforms


if __name__ == "__main__":
    test_170()
    test_170_2()
    test_170_3()



# -*- coding: utf-8 -*-
#
# Extra tests which are not part of the SHT or DASH test suites,
# nor the discrete issues tests or the cmdline_test file.
# The need for these tests are discovered by doing coverage checks and these
# are added as required.
import os
import re

from rdflib import Graph, Dataset

from pyshacl import validate
from pyshacl.errors import ReportableRuntimeError

ontology_graph_text = """
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .

<http://example.com/exOnt> a owl:Ontology ;
    rdfs:label "An example extra-ontology file."@en .

exOnt:Animal a rdfs:Class ;
    rdfs:comment "The parent class for Humans and Pets"@en ;
    rdfs:subClassOf owl:Thing .

exOnt:Human a rdfs:Class ;
    rdfs:comment "A Human being"@en ;
    rdfs:subClassOf exOnt:Animal .

exOnt:Pet a rdfs:Class ;
    rdfs:comment "An animal owned by a human"@en ;
    rdfs:subClassOf exOnt:Animal .

exOnt:hasPet a rdf:Property ;
    rdfs:domain exOnt:Human ;
    rdfs:range exOnt:Pet .

exOnt:nLegs a rdf:Property ;
    rdfs:domain exOnt:Animal ;
    rdfs:range xsd:integer .

exOnt:Teacher a rdfs:Class ;
    rdfs:comment "A Human who is a teacher."@en ;
    rdfs:subClassOf exOnt:Human .

exOnt:PreschoolTeacher a rdfs:Class ;
    rdfs:comment "A Teacher who teaches preschool."@en ;
    rdfs:subClassOf exOnt:Teacher .

exOnt:Lizard a rdfs:Class ;
    rdfs:subClassOf exOnt:Pet .

exOnt:Goanna a rdfs:Class ;
    rdfs:subClassOf exOnt:Lizard .

"""

ontology_ds_text = """
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .
# This is a TRIG file.

<http://example.com/exOnt-graph> {
<http://example.com/exOnt> a owl:Ontology ;
    rdfs:label "An example extra-ontology file."@en .

exOnt:Animal a rdfs:Class ;
    rdfs:comment "The parent class for Humans and Pets"@en ;
    rdfs:subClassOf owl:Thing .

exOnt:Human a rdfs:Class ;
    rdfs:comment "A Human being"@en ;
    rdfs:subClassOf exOnt:Animal .

exOnt:Pet a rdfs:Class ;
    rdfs:comment "An animal owned by a human"@en ;
    rdfs:subClassOf exOnt:Animal .

exOnt:hasPet a rdf:Property ;
    rdfs:domain exOnt:Human ;
    rdfs:range exOnt:Pet .

exOnt:nLegs a rdf:Property ;
    rdfs:domain exOnt:Animal ;
    rdfs:range xsd:integer .

exOnt:Teacher a rdfs:Class ;
    rdfs:comment "A Human who is a teacher."@en ;
    rdfs:subClassOf exOnt:Human .

exOnt:PreschoolTeacher a rdfs:Class ;
    rdfs:comment "A Teacher who teaches preschool."@en ;
    rdfs:subClassOf exOnt:Teacher .

exOnt:Lizard a rdfs:Class ;
    rdfs:subClassOf exOnt:Pet .

exOnt:Goanna a rdfs:Class ;
    rdfs:subClassOf exOnt:Lizard .
}
"""

shacl_file_text = """
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exShape: <http://example.com/exShape#> .
@prefix exOnt: <http://example.com/exOnt#> .

<http://example.com/exShape> a owl:Ontology ;
    rdfs:label "Example Shapes File"@en .

exShape:HumanShape a sh:NodeShape ;
    sh:property [
        sh:class exOnt:Pet ;
        sh:path exOnt:hasPet ;
    ] ;
    sh:property [
        sh:datatype xsd:integer ;
        sh:path exOnt:nLegs ;
        sh:maxInclusive 2 ;
        sh:minInclusive 2 ;
    ] ;
    sh:targetClass exOnt:Human .

exShape:AnimalShape a sh:NodeShape ;
    sh:property [
        sh:datatype xsd:integer ;
        sh:path exOnt:nLegs ;
        sh:maxInclusive 4 ;
        sh:minInclusive 1 ;
    ] ;
    sh:targetClass exOnt:Animal .
"""

data_file_text = """
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .
@prefix ex: <http://example.com/ex#> .

ex:Human1 rdf:type exOnt:PreschoolTeacher ;
    rdf:label "Amy" ;
    exOnt:nLegs "2"^^xsd:integer ;
    exOnt:hasPet ex:Pet1 .

ex:Pet1 rdf:type exOnt:Goanna ;
    rdf:label "Sebastian" ;
    exOnt:nLegs "4"^^xsd:integer .
"""

data_file_text_bad = """
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix exOnt: <http://example.com/exOnt#> .
@prefix ex: <http://example.com/ex#> .

ex:Human1 rdf:type exOnt:PreschoolTeacher ;
    rdf:label "Amy" ;
    exOnt:nLegs "2"^^xsd:integer ;
    exOnt:hasPet "Sebastian"^^xsd:string .

ex:Pet1 rdf:type exOnt:Goanna ;
    rdf:label "Sebastian" ;
    exOnt:nLegs "four"^^xsd:string .
"""


def test_validate_ds_with_graph_ontology():
    ds = Dataset()
    ds.parse(data=data_file_text_bad, format='turtle')
    extra_g = Graph()
    extra_g.parse(data=ontology_graph_text, format='turtle')

    ds_len = len(ds)
    res = validate(
        ds, shacl_graph=shacl_file_text, shacl_graph_format='turtle', ont_graph=extra_g, inference='rdfs', debug=True
    )
    conforms, graph, string = res
    assert not conforms
    # Assert that the dataset is unchanged
    ds_len2 = len(ds)
    assert ds_len2 == ds_len

def test_validate_ds_with_ds_ontology():
    ds = Dataset()
    ds.parse(data=data_file_text_bad, format='turtle')
    extra_ds = Dataset()
    extra_ds.parse(data=ontology_ds_text, format='trig')

    ds_len = len(ds)
    res = validate(
        ds, shacl_graph=shacl_file_text, shacl_graph_format='turtle', ont_graph=extra_ds, inference='rdfs', debug=True
    )
    conforms, graph, string = res
    assert not conforms
    # Assert that the dataset is unchanged
    ds_len2 = len(ds)
    assert ds_len2 == ds_len

def test_validate_ds_with_ds_ontology_inplace():
    ds = Dataset()
    ds.parse(data=data_file_text_bad, format='turtle')
    extra_ds = Dataset()
    extra_ds.parse(data=ontology_ds_text, format='trig')

    ds_len = len(ds)
    res = validate(
        ds,
        shacl_graph=shacl_file_text,
        shacl_graph_format='turtle',
        ont_graph=extra_ds,
        inference='rdfs',
        debug=True,
        inplace=True
    )
    conforms, graph, string = res
    assert not conforms
    # Assert that the dataset is changed
    ds_len2 = len(ds)
    assert ds_len2 != ds_len
    a = ds.serialize(format='trig')
    print(a)




# -*- coding: utf-8 -*-
#
import logging
from collections import OrderedDict, defaultdict
from os import path
from test.helpers import flatten_manifests, load_manifest

import pytest
from rdflib.namespace import RDF, RDFS, Namespace

import pyshacl
from pyshacl.errors import ReportableRuntimeError
from pyshacl.validator_conformance import check_sht_result

here_dir = path.abspath(path.dirname(__file__))
sht_files_dir = path.join(here_dir, 'resources', 'sht_tests')
sht_main_manifest = path.join(sht_files_dir, 'manifest.ttl')
MF = Namespace('http://www.w3.org/2001/sw/DataAccess/tests/test-manifest#')
SHT = Namespace('http://www.w3.org/ns/shacl-test#')

main_manifest = load_manifest(sht_main_manifest)
manifests_with_entries = flatten_manifests(main_manifest,True)

tests_found_in_manifests = defaultdict(lambda: [])

for m in manifests_with_entries:
    tests = m.collect_tests()
    tests_found_in_manifests[m.base].extend(tests)

tests_found_in_manifests = OrderedDict(sorted(tests_found_in_manifests.items()))
test_index_map = [[base, i] for base, tests in tests_found_in_manifests.items() for i, t in enumerate(tests)]

# There are some tests we know will fail, but we don't want to stop deployment
# if we hit them. List them here:
ALLOWABLE_FAILURES = ["/sparql/pre-binding/shapesGraph-001"]


@pytest.mark.parametrize("base, index", test_index_map)
def test_sht_all(base, index, caplog) -> None:
    caplog.set_level(logging.DEBUG)
    tests = tests_found_in_manifests[base]
    test = tests[index]
    run_sht_test(test, {"inference": 'rdfs', "debug": True, "meta_shacl": False})


@pytest.mark.parametrize("base, index", test_index_map)
def test_sht_all_sparql_mode(base, index, caplog) -> None:
    caplog.set_level(logging.DEBUG)
    tests = tests_found_in_manifests[base]
    test = tests[index]
    run_sht_test(test, {"inference": 'none', "debug": True, "sparql_mode": True, "meta_shacl": False})


def run_sht_test(sht_test, validate_args: dict) -> None:
    logger = logging.getLogger()  # pytest uses the root logger with a capturing handler
    if sht_test.node.startswith("file:///"):
        test_id = str(sht_test.node).replace("file:///", "")
    elif sht_test.node.startswith("file://"):
        test_id = str(sht_test.node).replace("file://", "")
    elif sht_test.node.startswith("file:"):
        test_id = str(sht_test.node).replace("file:", "")
    label = sht_test.label
    data_file = sht_test.data_graph
    shacl_file = sht_test.shapes_graph
    sparql_mode = validate_args.get('sparql_mode', False)
    if sparql_mode and shacl_file is None:
        # shacl_file cannot be None in SPARQL Remote Graph Mode
        shacl_file = data_file
    if label:
        logger.info("testing: ".format(label))
    try:
        conforms, r_graph, r_text = pyshacl.validate(data_file, shacl_graph=shacl_file, **validate_args)
    except (NotImplementedError, ReportableRuntimeError) as e:
        logger.exception(e)
        r_text = ""
        passes = False
    else:
        passes = check_sht_result(r_graph, sht_test.sht_graph, sht_test.sht_result, log=logger)
    logger.info(r_text)
    try:
        assert passes
    except AssertionError as ae:
        for af in ALLOWABLE_FAILURES:
            if test_id.endswith(af):
                logger.warning("Allowing failure in test: {}".format(test_id))
                break
        else:
            raise ae



from rdflib import Graph
from pyshacl import validate

shapes_graph = '''\
@prefix dash: <http://datashapes.org/dash#> .
@prefix ex: <http://datashapes.org/sh/tests/js/target/jsTargetType-001.test#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

<http://datashapes.org/sh/tests/js/target/jsTargetType-001.test>
  rdf:type owl:Ontology ;
  rdfs:label "Test of sh:JSTargetType 001" ;
  owl:imports <http://datashapes.org/dash> ;
.
ex:GraphValidationTestCase
  rdf:type dash:GraphValidationTestCase ;
  dash:expectedResult [
      rdf:type sh:ValidationReport ;
      sh:conforms "false"^^xsd:boolean ;
      sh:result [
          rdf:type sh:ValidationResult ;
          sh:focusNode ex:Barry ;
          sh:resultSeverity sh:Violation ;
          sh:sourceConstraintComponent sh:ClassConstraintComponent ;
          sh:sourceShape ex:USCitizenShape ;
          sh:value ex:Barry ;
        ] ;
    ] ;
.

ex:Person
  rdf:type owl:Class ;
  rdfs:label "A person" ;
.

ex:Country
  rdf:type owl:Class ;
  rdfs:label "A country" ;
.

ex:USA
  rdf:type ex:Country ;
.

ex:Germany
  rdf:type ex:Country ;
.

ex:bornIn
  rdf:type owl:ObjectProperty ;
.

ex:GermanCitizen
  rdf:type owl:Class ;
.
ex:USCitizen
  rdf:type owl:Class ;
.

ex:PeopleBornInCountryTarget
	a sh:JSTargetType ;
	rdfs:subClassOf sh:Target ;
	sh:labelTemplate "All persons born in {$country}" ;
	sh:parameter [
		sh:path ex:country ;
		sh:description "The country that the focus nodes are 'born' in." ;
		sh:class ex:Country ;
		sh:nodeKind sh:IRI ;
	] ;
	sh:jsFunctionName "findBornIn" ;
	sh:jsLibrary [ sh:jsLibraryURL "file:./test/resources/js/findBornIn.js"^^xsd:anyURI ] ;
.

ex:GermanCitizenShape
	a sh:NodeShape ;
	sh:target [
		a ex:PeopleBornInCountryTarget ;
		ex:country ex:Germany ;
	] ;
	sh:class ex:GermanCitizen ;
.

ex:USCitizenShape
	a sh:NodeShape ;
	sh:target [
		a ex:PeopleBornInCountryTarget ;
		ex:country ex:USA ;
	] ;
	sh:class ex:USCitizen ;
.
'''

data_graph = '''\
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix ex: <http://datashapes.org/sh/tests/js/target/jsTargetType-001.test#> .
@prefix exdata: <http://datashapes.org/sh/tests/js/target/jsTargetType-001.test.data#> .

exdata:Ludwig
  rdf:type ex:Person ;
  rdf:type ex:GermanCitizen ;
  ex:bornIn ex:Germany .

exdata:Barry
  rdf:type ex:Person ;
  ex:bornIn ex:USA .

'''


def test_js_target_type():
    s1 = Graph().parse(data=shapes_graph, format="turtle")
    g1 = Graph().parse(data=data_graph, format="turtle")
    conforms, result_graph, result_text = validate(g1, shacl_graph=s1, advanced=True, debug=True, js=True)
    assert not conforms


if __name__ == "__main__":
    test_js_target_type()



# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING, List, Optional, Sequence, Union

import rdflib
from rdflib import Literal
from rdflib.namespace import XSD

from pyshacl.consts import SH_construct
from pyshacl.errors import ReportableRuntimeError, RuleLoadError
from pyshacl.helper import get_query_helper_cls
from pyshacl.rdfutil import clone_graph

from ..shacl_rule import SHACLRule

if TYPE_CHECKING:
    from rdflib.term import URIRef

    from pyshacl.pytypes import GraphLike, RDFNode, SHACLExecutor
    from pyshacl.shape import Shape

XSD_string = XSD.string

SPARQL_RULE_ITERATE_LIMIT = 100


class SPARQLRule(SHACLRule):
    __slots__ = ("_constructs", "_qh")

    def __init__(self, executor: 'SHACLExecutor', shape: 'Shape', rule_node: 'rdflib.term.Identifier', **kwargs):
        """
        :param executor:
        :type executor: SHACLExecutor
        :param shape:
        :type shape: Shape
        :param rule_node:
        :type rule_node: rdflib.term.Identifier
        """
        super(SPARQLRule, self).__init__(executor, shape, rule_node, **kwargs)
        construct_nodes = set(self.shape.sg.objects(self.node, SH_construct))
        if len(construct_nodes) < 1:
            raise RuleLoadError("No sh:construct on SPARQLRule", "https://www.w3.org/TR/shacl-af/#SPARQLRule")
        self._constructs = []
        for c in construct_nodes:
            if not isinstance(c, Literal) or not (
                c.datatype == XSD_string or c.language is not None or isinstance(c.value, str)
            ):
                raise RuleLoadError(
                    "SPARQLRule sh:construct must be an xsd:string", "https://www.w3.org/TR/shacl-af/#SPARQLRule"
                )
            self._constructs.append(str(c.value))
        SPARQLQueryHelper = get_query_helper_cls()
        query_helper = SPARQLQueryHelper(self.shape, self.node, None, deactivated=self._deactivated)
        query_helper.collect_prefixes()
        self._qh = query_helper

    def apply(
        self,
        data_graph: 'GraphLike',
        focus_nodes: Optional[Sequence['RDFNode']] = None,
        target_graph_identifier: Optional['URIRef'] = None,
    ) -> int:
        focus_list: Sequence['RDFNode']
        if focus_nodes is not None:
            focus_list = list(focus_nodes)
        else:
            focus_list = list(self.shape.focus_nodes(data_graph))
        if self.executor.focus_nodes is not None and len(self.executor.focus_nodes) > 0:
            filtered_focus_nodes: List[Union[rdflib.URIRef]] = []
            for _fo in focus_list:  # type: RDFNode
                if isinstance(_fo, rdflib.URIRef) and _fo in self.executor.focus_nodes:
                    filtered_focus_nodes.append(_fo)
            len_filtered_focus = len(filtered_focus_nodes)
            if len_filtered_focus < 1:
                return 0
            focus_list = filtered_focus_nodes
        all_added = 0
        SPARQLQueryHelper = get_query_helper_cls()
        iterate_limit = int(SPARQL_RULE_ITERATE_LIMIT)
        while True:
            if iterate_limit < 1:
                raise ReportableRuntimeError(
                    f"Local SPARQLRule iteration exceeded iteration limit of {SPARQL_RULE_ITERATE_LIMIT}."
                )
            iterate_limit -= 1
            added = 0
            applicable_nodes = self.filter_conditions(focus_list, data_graph)
            construct_graphs = set()
            for a in applicable_nodes:
                for c in self._constructs:
                    init_bindings = {}
                    found_this = SPARQLQueryHelper.bind_this_regex.search(c)
                    if found_this:
                        init_bindings['this'] = a
                    c = self._qh.apply_prefixes(c)
                    results = data_graph.query(c, initBindings=init_bindings)
                    if results.type != "CONSTRUCT":
                        raise ReportableRuntimeError("Query executed by a SHACL SPARQLRule must be CONSTRUCT query.")
                    this_added = False
                    result_graph = results.graph
                    if result_graph is None:
                        raise ReportableRuntimeError("Query executed by a SHACL SPARQLRule did not return a Graph.")
                    for i in result_graph:
                        if not this_added and i not in data_graph:
                            this_added = True
                            # We only need to know at least one triple was added, then break!
                            break
                    if this_added:
                        added += 1
                        construct_graphs.add(result_graph)
            if added > 0:
                if isinstance(data_graph, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
                    if target_graph_identifier is not None:
                        target_graph = data_graph.get_context(target_graph_identifier)
                    else:
                        target_graph = data_graph.default_context
                else:
                    target_graph = data_graph
                for g in construct_graphs:
                    data_graph = clone_graph(g, target_graph=target_graph)
                all_added += added
                if self.iterate:
                    continue  # Jump up to iterate
                else:
                    break  # Don't iterate
            break  # We've reached a local steady state
        return all_added



# -*- coding: latin-1 -*-
#
from .entrypoints import shacl_rules, validate
from .rule_expand_runner import RuleExpandRunner
from .shape import Shape
from .shapes_graph import ShapesGraph
from .validator import Validator

# version compliant with https://www.python.org/dev/peps/pep-0440/
__version__ = '0.30.0'
# Don't forget to change the version number in pyproject.toml, Dockerfile, and CITATION.cff along with this one

__all__ = ['validate', 'shacl_rules', 'Validator', 'RuleExpandRunner', '__version__', 'Shape', 'ShapesGraph']



# -*- coding: utf-8 -*-
#

from dataclasses import dataclass
from typing import List, Optional, Union

from rdflib import ConjunctiveGraph, Dataset, Graph, Literal
from rdflib.term import IdentifiedNode, URIRef

ConjunctiveLike = Union[ConjunctiveGraph, Dataset]
GraphLike = Union[ConjunctiveLike, Graph]
RDFNode = Union[IdentifiedNode, Literal]


@dataclass
class SHACLExecutor:
    validator: Optional[object] = None
    advanced_mode: bool = False
    abort_on_first: bool = False
    allow_infos: bool = False
    allow_warnings: bool = False
    iterate_rules: bool = False
    debug: bool = False
    sparql_mode: bool = False
    max_validation_depth: int = 15
    focus_nodes: Optional[List[URIRef]] = None



# -*- coding: utf-8 -*-
"""
https://www.w3.org/TR/shacl/#core-components-shape
"""

from textwrap import indent
from typing import Dict, List, Optional
from warnings import warn

import rdflib

from pyshacl.constraints.constraint_component import ConstraintComponent
from pyshacl.consts import (
    SH,
    SH_detail,
    SH_node,
    SH_NodeConstraintComponent,
    SH_property,
    SH_PropertyConstraintComponent,
)
from pyshacl.errors import (
    ConstraintLoadError,
    ConstraintLoadWarning,
    ReportableRuntimeError,
    ShapeRecursionWarning,
    ValidationFailure,
)
from pyshacl.pytypes import GraphLike, SHACLExecutor
from pyshacl.rdfutil import stringify_node
from pyshacl.shape import Shape

SH_QualifiedValueCountConstraintComponent = SH.QualifiedValueConstraintComponent
SH_QualifiedMaxCountConstraintComponent = SH.QualifiedMaxCountConstraintComponent
SH_QualifiedMinCountConstraintComponent = SH.QualifiedMinCountConstraintComponent

SH_qualifiedValueShape = SH.qualifiedValueShape
SH_qualifiedValueShapesDisjoint = SH.qualifiedValueShapesDisjoint
SH_qualifiedMinCount = SH.qualifiedMinCount
SH_qualifiedMaxCount = SH.qualifiedMaxCount


class PropertyConstraintComponent(ConstraintComponent):
    """
    sh:property can be used to specify that each value node has a given property shape.
    Link:
    https://www.w3.org/TR/shacl/#PropertyConstraintComponent
    Textual Definition:
    For each value node v: A failure MUST be produced if the validation of v as focus node against the property shape $property produces a failure. Otherwise, the validation results are the results of validating v as focus node against the property shape $property.
    """

    shacl_constraint_component = SH_PropertyConstraintComponent
    shape_expecting = True
    list_taking = False

    def __init__(self, shape: Shape) -> None:
        super(PropertyConstraintComponent, self).__init__(shape)
        property_shapes = list(self.shape.objects(SH_property))
        if len(property_shapes) < 1:
            raise ConstraintLoadError(
                "PropertyConstraintComponent must have at least one sh:property predicate.",
                "https://www.w3.org/TR/shacl/#PropertyConstraintComponent",
            )
        self.property_shapes = property_shapes

    @classmethod
    def constraint_parameters(cls) -> List[rdflib.URIRef]:
        return [SH_property]

    @classmethod
    def constraint_name(cls) -> str:
        return "PropertyConstraintComponent"

    def make_generic_messages(self, datagraph: GraphLike, focus_node, value_node) -> List[rdflib.Literal]:
        raise NotImplementedError("A Property Constraint Component should not be able to generate its own message.")

    def evaluate(
        self, executor: SHACLExecutor, target_graph: GraphLike, focus_value_nodes: Dict, _evaluation_path: List
    ):
        """
        Entrypoint for constraint evaluation.
        :type executor: SHACLExecutor
        :type target_graph: rdflib.Graph
        :type focus_value_nodes: dict
        :type _evaluation_path: list
        """
        reports: List[Dict] = []
        non_conformant = False

        # Shortcut, when there are no value nodes, don't check for recursion, don't validate and exit early
        value_node_count = 0
        for f, value_nodes in focus_value_nodes.items():
            value_node_count = value_node_count + len(value_nodes)
        if value_node_count < 1:
            return (not non_conformant), reports

        potentially_recursive = self.recursion_triggers(_evaluation_path)

        for p_shape in self.property_shapes:
            _nc, _r = self._evaluate_property_shape(
                executor, p_shape, target_graph, focus_value_nodes, potentially_recursive, _evaluation_path
            )
            non_conformant = non_conformant or _nc
            reports.extend(_r)
        return (not non_conformant), reports

    def _evaluate_property_shape(
        self, executor, prop_shape, target_graph, focus_value_nodes, potentially_recursive, _evaluation_path
    ):
        _reports = []
        _non_conformant = False
        prop_shape = self.shape.get_other_shape(prop_shape)
        if potentially_recursive and prop_shape in potentially_recursive:
            warn(ShapeRecursionWarning(_evaluation_path))
            return _non_conformant, _reports
        if not prop_shape:
            raise ReportableRuntimeError(
                f"SHACL PropertyShape not found: The shape referenced by sh:property does not exist. "
                f"Please check if the shape '{prop_shape}' is defined."
            )
        elif not prop_shape.is_property_shape:
            raise ReportableRuntimeError(
                f"'{prop_shape}' exists but is not a well-formed SHACL PropertyShape. "
                f"Ensure it has the correct type (sh:PropertyShape) and all required properties."
            )

        for f, value_nodes in focus_value_nodes.items():
            for v in value_nodes:
                _is_conform, _r = prop_shape.validate(
                    executor, target_graph, focus=v, _evaluation_path=_evaluation_path[:]
                )
                _non_conformant = _non_conformant or (not _is_conform)
                _reports.extend(_r)
        return _non_conformant, _reports


class NodeConstraintComponent(ConstraintComponent):
    """
    sh:node specifies the condition that each value node conforms to the given node shape.
    Link:
    https://www.w3.org/TR/shacl/#NodeShapeComponent
    Textual Definition:
    For each value node v: A failure MUST be produced if the conformance checking of v against $node produces a failure. Otherwise, if v does not conform to $node, there is a validation result with v as sh:value.
    """

    shacl_constraint_component = SH_NodeConstraintComponent
    shape_expecting = True
    list_taking = False

    def __init__(self, shape: Shape) -> None:
        super(NodeConstraintComponent, self).__init__(shape)
        node_shapes = list(self.shape.objects(SH_node))
        if len(node_shapes) < 1:
            raise ConstraintLoadError(
                "NodeConstraintComponent must have at least one sh:node predicate.",
                "https://www.w3.org/TR/shacl/#NodeConstraintComponent",
            )
        self.node_shapes = node_shapes

    @classmethod
    def constraint_parameters(cls) -> List[rdflib.URIRef]:
        return [SH_node]

    @classmethod
    def constraint_name(cls) -> str:
        return "NodeConstraintComponent"

    def make_generic_messages(self, datagraph: GraphLike, focus_node, value_node) -> List[rdflib.Literal]:
        if len(self.node_shapes) < 2:
            m = "Value does not conform to Shape {}.".format(stringify_node(self.shape.sg.graph, self.node_shapes[0]))
        else:
            rules = "', '".join(stringify_node(self.shape.sg.graph, c) for c in self.node_shapes)
            m = "Value must conform to every Shape in ('{}').".format(rules)
        m += " See details for more information."
        return [rdflib.Literal(m)]

    def evaluate(
        self, executor: SHACLExecutor, target_graph: GraphLike, focus_value_nodes: Dict, _evaluation_path: List
    ):
        """
        :type executor: SHACLExecutor
        :type target_graph: rdflib.Graph
        :type focus_value_nodes: dict
        :type _evaluation_path: list
        """
        reports: List[Dict] = []
        non_conformant = False

        # Shortcut, when there are no value nodes, don't check for recursion, don't validate and exit early
        value_node_count = 0
        for f, value_nodes in focus_value_nodes.items():
            value_node_count = value_node_count + len(value_nodes)
        if value_node_count < 1:
            return (not non_conformant), reports

        potentially_recursive = self.recursion_triggers(_evaluation_path)

        for n_shape in self.node_shapes:
            _nc, _r = self._evaluate_node_shape(
                executor, n_shape, target_graph, focus_value_nodes, potentially_recursive, _evaluation_path
            )
            non_conformant = non_conformant or _nc
            reports.extend(_r)
        return (not non_conformant), reports

    def _evaluate_node_shape(
        self, executor, node_shape, target_graph, focus_value_nodes, potentially_recursive, _evaluation_path
    ):
        _reports = []
        _non_conformant = False
        node_shape = self.shape.get_other_shape(node_shape)
        if potentially_recursive and node_shape in potentially_recursive:
            warn(ShapeRecursionWarning(_evaluation_path))
            return _non_conformant, _reports
        if not node_shape or node_shape.is_property_shape:
            raise ReportableRuntimeError(
                "Shape pointed to by sh:node does not exist or is not a well-formed SHACL NodeShape."
            )
        for f, value_nodes in focus_value_nodes.items():
            for v in value_nodes:
                _is_conform, _r = node_shape.validate(
                    executor, target_graph, focus=v, _evaluation_path=_evaluation_path[:]
                )
                # Create a failure for this constraint component if any failures exist
                if (not _is_conform) or len(_r) > 0:
                    _non_conformant = True
                    rept_text, rept_node, rept_triples = self.make_v_result(target_graph, f, value_node=v)
                    # Nest the others underneath via sh:detail
                    rept_text = f"{rept_text}\tDetails:\n"
                    for text_sub, node_sub, triples_sub in _r:
                        # Add text of validation result in nested details section
                        rept_text += indent(text_sub, "\t\t")
                        # Add a triple connecting the new validation result to the
                        # validation result for the nested node
                        rept_triples.append((rept_node, SH_detail, node_sub))
                        # Extend the triples in the report with the ones from the nested result
                        rept_triples.extend(triples_sub)
                    _reports.append((rept_text, rept_node, rept_triples))
        return _non_conformant, _reports


class QualifiedValueShapeConstraintComponent(ConstraintComponent):
    """
    sh:qualifiedValueShape specifies the condition that a specified number of value nodes conforms to the given shape. Each sh:qualifiedValueShape can have: one value for sh:qualifiedMinCount, one value for sh:qualifiedMaxCount or, one value for each, at the same subject.
    Link:
    https://www.w3.org/TR/shacl/#QualifiedValueShapeConstraintComponent
    Textual Definition for qualifiedMinCount:
    Let C be the number of value nodes v where v conforms to $qualifiedValueShape and where v does not conform to any of the sibling shapes for the current shape, i.e. the shape that v is validated against and which has $qualifiedValueShape as its value for sh:qualifiedValueShape. A failure MUST be produced if any of the said conformance checks produces a failure. Otherwise, there is a validation result if C is less than $qualifiedMinCount. The constraint component for sh:qualifiedMinCount is sh:QualifiedMinCountConstraintComponent.
    Textual Definition for qualifiedMaxCount:
    Let C be as defined for sh:qualifiedMinCount above. A failure MUST be produced if any of the said conformance checks produces a failure. Otherwise, there is a validation result if C is greater than $qualifiedMaxCount. The constraint component for sh:qualifiedMaxCount is sh:QualifiedMaxCountConstraintComponent.
    """

    shacl_constraint_component = NotImplemented
    shape_expecting = True
    list_taking = False

    def __init__(self, shape: Shape) -> None:
        super(QualifiedValueShapeConstraintComponent, self).__init__(shape)
        if not shape.is_property_shape:
            # Note, this no longer throws an error, this constraint is simply ignored on NodeShapes.
            raise ConstraintLoadWarning(
                "QualifiedValueShapeConstraintComponent can only be present on a PropertyShape, not a NodeShape.",
                "https://www.w3.org/TR/shacl/#QualifiedValueShapeConstraintComponent",
            )
        value_shapes = set(self.shape.objects(SH_qualifiedValueShape))
        if len(value_shapes) < 1:
            raise ConstraintLoadError(
                "QualifiedValueShapeConstraintComponent must have at least one sh:qualifiedValueShape predicate.",
                "https://www.w3.org/TR/shacl/#QualifiedValueShapeConstraintComponent",
            )
        self.value_shapes = value_shapes
        min_count: Optional[int]
        min_counts = set(self.shape.objects(SH_qualifiedMinCount))
        if len(min_counts) < 1:
            min_count = None
        elif len(min_counts) > 1:
            raise ConstraintLoadError(
                "QualifiedMinCountConstraintComponent must have at most one sh:qualifiedMinCount predicate.",
                "https://www.w3.org/TR/shacl/#QualifiedValueShapeConstraintComponent",
            )
        else:
            min_count_literal = next(iter(min_counts))
            if not isinstance(min_count_literal, rdflib.Literal) or not isinstance(min_count_literal.value, int):
                raise ConstraintLoadError(
                    "QualifiedMinCountConstraintComponent sh:qualifiedMinCount must be a Literal with Int.",
                    "https://www.w3.org/TR/shacl/#QualifiedValueShapeConstraintComponent",
                )
            min_count = min_count_literal.value

        max_count: Optional[int]
        max_counts = set(self.shape.objects(SH_qualifiedMaxCount))
        if len(max_counts) < 1:
            max_count = None
        elif len(max_counts) > 1:
            raise ConstraintLoadError(
                "QualifiedMaxCountConstraintComponent must have at most one sh:qualifiedMaxCount predicate.",
                "https://www.w3.org/TR/shacl/#QualifiedValueShapeConstraintComponent",
            )
        else:
            max_count_literal = next(iter(max_counts))
            if not isinstance(max_count_literal, rdflib.Literal) or not isinstance(max_count_literal.value, int):
                raise ConstraintLoadError(
                    "QualifiedMaxCountConstraintComponent sh:qualifiedMaxCount must be a Literal with Int.",
                    "https://www.w3.org/TR/shacl/#QualifiedValueShapeConstraintComponent",
                )
            max_count = max_count_literal.value

        if min_count is None and max_count is None:
            raise ConstraintLoadError(
                "QualifiedValueShapeConstraintComponent must have at lease one sh:qualifiedMinCount or "
                "sh:qualifiedMaxCount",
                "https://www.w3.org/TR/shacl/#QualifiedValueShapeConstraintComponent",
            )
        is_disjoint = False
        disjoint_nodes = set(self.shape.objects(SH_qualifiedValueShapesDisjoint))
        for d in disjoint_nodes:
            if isinstance(d, rdflib.Literal):
                if isinstance(d.value, bool):
                    is_disjoint = is_disjoint or d.value
        self.min_count = min_count
        self.max_count = max_count
        self.is_disjoint = is_disjoint

    @classmethod
    def constraint_parameters(cls) -> List[rdflib.URIRef]:
        return [SH_qualifiedValueShape, SH_qualifiedMinCount, SH_qualifiedValueShapesDisjoint, SH_qualifiedMaxCount]

    @classmethod
    def constraint_name(cls) -> str:
        return "QualifiedValueShapeConstraintComponent"

    def make_generic_messages(self, datagraph: GraphLike, focus_node, value_node) -> List[rdflib.Literal]:
        # TODO:
        #  Implement default message for QualifiedValueConstraint (seems messy)
        shapes_string = ",".join(stringify_node(self.shape.sg.graph, s) for s in self.value_shapes)
        count_message = ""
        if self.min_count is not None:
            count_message += f" MinCount {self.min_count}"
        if self.max_count is not None:
            count_message += f" MaxCount {self.max_count}"
        if len(self.value_shapes) > 1:
            return [rdflib.Literal(f"Focus node does not conform to shapes{count_message}: ({shapes_string})")]
        return [rdflib.Literal(f"Focus node does not conform to shape{count_message}: {shapes_string}")]

    def evaluate(
        self, executor: SHACLExecutor, target_graph: GraphLike, focus_value_nodes: Dict, _evaluation_path: List
    ):
        """
        :type executor: SHACLExecutor
        :type target_graph: rdflib.Graph
        :type focus_value_nodes: dict
        :type _evaluation_path: list
        """
        reports: List[Dict] = []
        non_conformant = False

        # Shortcut, when there are no value nodes, don't check for recursion, don't validate and exit early
        value_node_count = 0
        for f, value_nodes in focus_value_nodes.items():
            value_node_count = value_node_count + len(value_nodes)
        if not self.is_disjoint and value_node_count < 1 and (self.min_count is None or self.min_count < 1):
            return (not non_conformant), reports

        potentially_recursive = self.recursion_triggers(_evaluation_path)

        for v_shape in self.value_shapes:
            _nc, _r = self._evaluate_value_shape(
                executor, v_shape, target_graph, focus_value_nodes, potentially_recursive, _evaluation_path
            )
            non_conformant = non_conformant or _nc
            reports.extend(_r)
        return (not non_conformant), reports

    def _evaluate_value_shape(
        self, executor, _v_shape, target_graph, focus_value_nodes, potentially_recursive, _evaluation_path
    ):
        _reports = []
        _non_conformant = False
        other_shape = self.shape.get_other_shape(_v_shape)
        if potentially_recursive and other_shape in potentially_recursive:
            warn(ShapeRecursionWarning(_evaluation_path))
            return _non_conformant, _reports
        if not other_shape:
            raise ReportableRuntimeError(
                "Shape pointed to by sh:qualifiedValueShape does not exist or is not a well-formed SHACL Shape."
            )
        if self.is_disjoint:
            # Textual Definition of Sibling Shapes:
            # Let Q be a shape in shapes graph G that declares a qualified cardinality constraint (by having values for sh:qualifiedValueShape and at least one of sh:qualifiedMinCount or sh:qualifiedMaxCount). Let ps be the set of shapes in G that have Q as a value of sh:property. If Q has true as a value for sh:qualifiedValueShapesDisjoint then the set of sibling shapes for Q is defined as the set of all values of the SPARQL property path sh:property/sh:qualifiedValueShape for any shape in ps minus the value of sh:qualifiedValueShape of Q itself. The set of sibling shapes is empty otherwise.
            sibling_shapes = set()
            parent_shapes = set(self.shape.sg.subjects(SH_property, self.shape.node))
            for p in iter(parent_shapes):
                parent_property_shapes = set(self.shape.sg.objects(p, SH_property))
                for s in iter(parent_property_shapes):
                    parent_property_qualifiedvalueshapes = set(self.shape.sg.objects(s, SH_qualifiedValueShape))
                    for sibling in parent_property_qualifiedvalueshapes:
                        if sibling == _v_shape:
                            continue
                        sibling_shapes.add(sibling)

            sibling_shapes = set(self.shape.get_other_shape(s) for s in sibling_shapes)
        else:
            sibling_shapes = set()
        upstream_reports = []
        for f, value_nodes in focus_value_nodes.items():
            number_conforms = 0
            for v in value_nodes:
                try:
                    _is_conform, _r = other_shape.validate(
                        executor, target_graph, focus=v, _evaluation_path=_evaluation_path[:]
                    )
                    if len(_r):
                        upstream_reports.extend(_r)
                    if _is_conform:
                        _conforms_to_sibling = False
                        for sibling_shape in sibling_shapes:
                            _c2, _r = sibling_shape.validate(
                                executor, target_graph, focus=v, _evaluation_path=_evaluation_path[:]
                            )
                            _conforms_to_sibling = _conforms_to_sibling or _c2
                        if not _conforms_to_sibling:
                            number_conforms += 1
                except ValidationFailure as v:
                    raise v
            if self.max_count is not None and number_conforms > self.max_count:
                _non_conformant = True
                _r = self.make_v_result(target_graph, f, constraint_component=SH_QualifiedMaxCountConstraintComponent)
                _reports.append(_r)
            if self.min_count is not None and number_conforms < self.min_count:
                _non_conformant = True
                _r = self.make_v_result(target_graph, f, constraint_component=SH_QualifiedMinCountConstraintComponent)
                _reports.append(_r)
        if len(upstream_reports) and self.shape.sg.debug:
            self.shape.logger.debug(
                "sh:qualifiedValueShape constraint reports will be ignored and not passed to the parent Node:"
            )
            for v_str, v_node, v_parts in upstream_reports:
                self.shape.logger.debug(v_str)
        return _non_conformant, _reports



# -*- coding: utf-8 -*-
#
"""
SHACL-AF Advanced Constraints
https://www.w3.org/TR/shacl-af/#ExpressionConstraintComponent
"""

import typing
from typing import Dict, List

from rdflib import Literal, URIRef

from pyshacl.constraints.constraint_component import ConstraintComponent
from pyshacl.consts import SH, SH_message
from pyshacl.errors import ConstraintLoadError
from pyshacl.helper.expression_helper import nodes_from_node_expression
from pyshacl.pytypes import GraphLike, SHACLExecutor

SH_expression = SH.expression
SH_ExpressionConstraintComponent = SH.ExpressionConstraintComponent

if typing.TYPE_CHECKING:
    from pyshacl.shape import Shape


class ExpressionConstraint(ConstraintComponent):
    shacl_constraint_component = SH_ExpressionConstraintComponent

    def __init__(self, shape: 'Shape') -> None:
        super(ExpressionConstraint, self).__init__(shape)
        self.expr_nodes = list(self.shape.objects(SH_expression))
        if len(self.expr_nodes) < 1:
            raise ConstraintLoadError(
                "ExpressionConstraintComponent must have at least one sh:expression predicate.",
                "https://www.w3.org/TR/shacl-af/#ExpressionConstraintComponent",
            )

    @classmethod
    def constraint_parameters(cls) -> List[URIRef]:
        return [SH_expression]

    @classmethod
    def constraint_name(cls) -> str:
        return "ExpressionConstraintComponent"

    def make_generic_messages(self, datagraph: GraphLike, focus_node, value_node) -> List[Literal]:
        return [Literal("Expression evaluation generated constraint did not return true.")]

    def evaluate(
        self, executor: SHACLExecutor, data_graph: GraphLike, focus_value_nodes: Dict, _evaluation_path: List
    ):
        """
        :type executor: SHACLExecutor
        :type data_graph: rdflib.Graph
        :type focus_value_nodes: dict
        :type _evaluation_path: list
        """
        reports = []
        non_conformant = False
        for n in self.expr_nodes:
            _n, _r = self._evaluate_expression(data_graph, focus_value_nodes, n)
            non_conformant = non_conformant or _n
            reports.extend(_r)
        return (not non_conformant), reports

    def _evaluate_expression(self, data_graph, f_v_dict, expr):
        reports = []
        non_conformant = False
        messages = list(self.shape.sg.objects(expr, SH_message))
        if len(messages):
            messages = [next(iter(messages))]
        else:
            messages = None
        for f, value_nodes in f_v_dict.items():
            for v in value_nodes:
                try:
                    n_set = nodes_from_node_expression(expr, v, data_graph, self.shape.sg)
                    if (
                        isinstance(n_set, (list, set))
                        and len(n_set) == 1
                        and next(iter(n_set)) in (Literal(True), True)
                    ):
                        ...
                    else:
                        non_conformant = non_conformant or True
                        reports.append(
                            self.make_v_result(
                                data_graph, f, value_node=v, source_constraint=expr, extra_messages=messages
                            )
                        )
                except Exception as e:
                    print(e)
                    raise
        return non_conformant, reports



# -*- coding: utf-8 -*-
#
import itertools
import logging
import sys
from decimal import Decimal
from time import perf_counter
from typing import TYPE_CHECKING, Dict, List, Optional, Sequence, Set, Type, Union

from rdflib import BNode, IdentifiedNode, Literal, URIRef

from .consts import (
    RDF_type,
    RDFS_Class,
    RDFS_subClassOf,
    SH_deactivated,
    SH_description,
    SH_Info,
    SH_jsFunctionName,
    SH_JSTarget,
    SH_JSTargetType,
    SH_message,
    SH_name,
    SH_order,
    SH_property,
    SH_resultSeverity,
    SH_select,
    SH_severity,
    SH_SPARQLTarget,
    SH_SPARQLTargetType,
    SH_target,
    SH_targetClass,
    SH_targetNode,
    SH_targetObjectsOf,
    SH_targetSubjectsOf,
    SH_Violation,
    SH_Warning,
)
from .errors import ConstraintLoadError, ConstraintLoadWarning, ReportableRuntimeError, ShapeLoadError
from .helper import get_query_helper_cls
from .helper.expression_helper import value_nodes_from_path
from .helper.path_helper import shacl_path_to_sparql_path
from .pytypes import GraphLike, RDFNode, SHACLExecutor

if TYPE_CHECKING:
    from pyshacl.constraints import ConstraintComponent
    from pyshacl.shapes_graph import ShapesGraph

module = sys.modules[__name__]


class Shape(object):
    __slots__ = (
        'logger',
        'sg',
        'node',
        '_p',
        '_path',
        '_advanced',
        '_deactivated',
        '_severity',
        '_messages',
        '_names',
        '_descriptions',
    )

    def __init__(
        self,
        sg: 'ShapesGraph',
        node: Union[URIRef, BNode],
        p=False,
        path: Optional[Union[URIRef, BNode]] = None,
        logger=None,
    ):
        """
        Shape
        :type sg: ShapesGraph
        :type node: URIRef | BNode
        :type p: bool
        :type path: URIRef | BNode | None
        :type logger: logging.Logger
        """
        self.logger = logger or logging.getLogger(__name__)
        self.sg = sg
        self.node = node
        self._p = p
        self._path = path
        self._advanced = False

        deactivated_vals = set(self.objects(SH_deactivated))
        if len(deactivated_vals) > 1:
            # TODO:coverage: we don't have any tests for invalid shapes
            raise ShapeLoadError(
                "A SHACL Shape cannot have more than one sh:deactivated predicate.",
                "https://www.w3.org/TR/shacl/#deactivated",
            )
        elif len(deactivated_vals) < 1:
            self._deactivated = False  # type: bool
        else:
            d = next(iter(deactivated_vals))
            if not isinstance(d, Literal):
                # TODO:coverage: we don't have any tests for invalid shapes
                raise ShapeLoadError(
                    "The value of sh:deactivated predicate on a SHACL Shape must be a Literal.",
                    "https://www.w3.org/TR/shacl/#deactivated",
                )
            self._deactivated = bool(d.value)
        severity = set(self.objects(SH_severity))
        if len(severity):
            self._severity = next(iter(severity))  # type: Union[URIRef, BNode, Literal]
        else:
            self._severity = SH_Violation
        messages = set(self.objects(SH_message))
        if len(messages):
            self._messages = messages  # type: Set
        else:
            self._messages = set()
        names = set(self.objects(SH_name))
        if len(names):
            self._names = names  # type: Set
        else:
            self._names = set()
        descriptions = set(self.objects(SH_description))
        if len(descriptions):
            self._descriptions = descriptions  # type: Set
        else:
            self._descriptions = set()

    def set_advanced(self, val):
        self._advanced = bool(val)

    def get_other_shape(self, shape_node):
        try:
            return self.sg.lookup_shape_from_node(shape_node)
        except (KeyError, AttributeError):
            # TODO:coverage: we never hit this during a successful test run
            return None

    @property
    def is_property_shape(self):
        return bool(self._p)

    def property_shapes(self):
        # TODO:coverage: this is never used?
        return self.sg.graph.objects(self.node, SH_property)

    @property
    def deactivated(self):
        return self._deactivated

    @property
    def severity(self):
        return self._severity

    @property
    def message(self):
        if self._messages is None:
            return
        for m in self._messages:
            yield m

    @property
    def name(self):
        if self._names is None:
            return
        for n in self._names:
            yield n

    def __str__(self):
        try:
            name = next(iter(self.name))
        except Exception:
            name = str(self.node)
        if self.is_property_shape:
            kind = "PropertyShape"
        else:
            kind = "NodeShape"
        return "<{} {}>".format(kind, name)

    def __repr__(self):
        if self.is_property_shape:
            p = "True"
        else:
            p = "False"
        names = list(self.name)
        if len(names):
            return "<Shape {} p={} node={}>".format(",".join(names), p, str(self.node))
        else:
            return "<Shape p={} node={}>".format(p, str(self.node))
        # return super(Shape, self).__repr__()

    @property
    def description(self):
        # TODO:coverage: this is never used?
        if self._descriptions is None:
            return
        for d in self._descriptions:
            yield d

    def objects(self, predicate=None):
        return self.sg.graph.objects(self.node, predicate)

    @property
    def order(self):
        order_nodes = list(self.objects(SH_order))
        if len(order_nodes) < 1:
            return Decimal("0.0")
        if len(order_nodes) > 1:
            raise ShapeLoadError(
                "A SHACL Shape can have only one sh:order property.", "https://www.w3.org/TR/shacl-af/#rules-order"
            )
        order_node = next(iter(order_nodes))
        if not isinstance(order_node, Literal):
            raise ShapeLoadError(
                "A SHACL Shape must be a numeric literal.", "https://www.w3.org/TR/shacl-af/#rules-order"
            )
        if isinstance(order_node.value, Decimal):
            order = order_node.value
        elif isinstance(order_node.value, int):
            order = Decimal(order_node.value)
        elif isinstance(order_node.value, float):
            order = Decimal(str(order_node.value))
        else:
            raise ShapeLoadError(
                "A SHACL Shape must be a numeric literal.", "https://www.w3.org/TR/shacl-af/#rules-order"
            )
        return order

    def target_nodes(self):
        return self.sg.graph.objects(self.node, SH_targetNode)

    def target_classes(self):
        return self.sg.graph.objects(self.node, SH_targetClass)

    def implicit_class_targets(self):
        types = list(self.sg.graph.objects(self.node, RDF_type))
        subclasses = list(self.sg.graph.subjects(RDFS_subClassOf, RDFS_Class))
        subclasses.append(RDFS_Class)
        for t in types:
            if t in subclasses:
                return [self.node]
        return []

    def target_objects_of(self):
        return self.sg.graph.objects(self.node, SH_targetObjectsOf)

    def target_subjects_of(self):
        return self.sg.graph.objects(self.node, SH_targetSubjectsOf)

    def path(self):
        if not self.is_property_shape:
            return None
        if self._path is not None:
            return self._path
        raise RuntimeError("property shape has no _path!")  # pragma: no cover

    def target(self):
        target_nodes = self.target_nodes()
        target_classes = self.target_classes()
        implicit_targets = self.implicit_class_targets()
        target_objects_of = self.target_objects_of()
        target_subjects_of = self.target_subjects_of()
        return (target_nodes, target_classes, implicit_targets, target_objects_of, target_subjects_of)

    def advanced_target(self):
        custom_targets = set(self.sg.objects(self.node, SH_target))
        result_set = dict()
        if self.sg.js_enabled:
            use_JSTarget: Union[bool, Type] = True
        else:
            use_JSTarget = False

        for c in custom_targets:
            ct = dict()
            selects = list(self.sg.objects(c, SH_select))
            has_select = len(selects) > 0
            fn_names = list(self.sg.objects(c, SH_jsFunctionName))
            has_fnname = len(fn_names) > 0
            is_types = set(self.sg.objects(c, RDF_type))
            if has_select or (SH_SPARQLTarget in is_types):
                ct['type'] = SH_SPARQLTarget
                SPARQLQueryHelper = get_query_helper_cls()
                qh = SPARQLQueryHelper(self, c, selects[0], deactivated=self._deactivated)
                qh.collect_prefixes()
                ct['qh'] = qh
            elif has_fnname or (SH_JSTarget in is_types):
                if use_JSTarget:
                    JST = getattr(module, "JSTarget", None)
                    if not JST:
                        # Lazy-import JS-Target to prevent RDFLib import error
                        from pyshacl.extras.js.target import JSTarget as JST

                        setattr(module, "JSTarget", JST)
                    ct['type'] = SH_JSTarget
                    ct['targeter'] = JST(self.sg, c)
                else:
                    #  Found JSTarget, but JS is not enabled in PySHACL. Ignore this target.
                    pass
            else:
                found_tt = None
                for t in is_types:
                    try:
                        found_tt = self.sg.get_shacl_target_type(t)
                        break
                    except LookupError:
                        continue
                if not found_tt:
                    msg = "None of these types match a TargetType: {}".format(" ".join(is_types))
                    raise ShapeLoadError(msg, "https://www.w3.org/TR/shacl-af/#SPARQLTargetType")
                bound_tt = found_tt.bind(self, c)
                ct['type'] = bound_tt.shacl_constraint_class()
                if ct['type'] == SH_SPARQLTargetType:
                    ct['qt'] = bound_tt
                elif ct['type'] == SH_JSTargetType:
                    ct['targeter'] = bound_tt
            result_set[c] = ct
        return result_set

    def focus_nodes(self, data_graph, debug=False):
        """
        The set of focus nodes for a shape may be identified as follows:

        specified in a shape using target declarations
        specified in any constraint that references a shape in parameters of shape-expecting constraint parameters (e.g. sh:node)
        specified as explicit input to the SHACL processor for validating a specific RDF term against a shape
        :return:
        """
        t1 = 0.0
        if debug:
            t1 = perf_counter()
        (target_nodes, target_classes, implicit_classes, target_objects_of, target_subjects_of) = self.target()
        if self._advanced:
            advanced_targets = self.advanced_target()
        else:
            advanced_targets = False
        found_node_targets = set()
        # Just add _all_ target_nodes to the set,
        # they don't need to actually exist in the graph
        found_node_targets.update(iter(target_nodes))
        target_classes = set(target_classes)
        target_classes.update(set(implicit_classes))
        found_target_instances = set()
        for tc in target_classes:
            s = data_graph.subjects(RDF_type, tc)
            found_target_instances.update(s)
            subc = data_graph.transitive_subjects(RDFS_subClassOf, tc)
            for subclass in iter(subc):
                if subclass == tc:
                    continue
                s1 = data_graph.subjects(RDF_type, subclass)
                found_target_instances.update(s1)
        found_node_targets.update(found_target_instances)
        found_target_subject_of = set()
        for s_of in target_subjects_of:
            subs = {s for s, o in data_graph.subject_objects(s_of)}
            found_target_subject_of.update(subs)
        found_node_targets.update(found_target_subject_of)
        found_target_object_of = set()
        for o_of in target_objects_of:
            objs = {o for s, o in data_graph.subject_objects(o_of)}
            found_target_object_of.update(objs)
        found_node_targets.update(found_target_object_of)
        if advanced_targets:
            for at_node, at in advanced_targets.items():
                if at['type'] == SH_SPARQLTarget:
                    qh = at['qh']
                    select = qh.apply_prefixes(qh.select_text)
                    results = data_graph.query(select, initBindings=None)
                    if not results or len(results.bindings) < 1:
                        continue
                    for r in results:
                        t = r['this']
                        found_node_targets.add(t)
                elif at['type'] in (SH_JSTarget, SH_JSTargetType):
                    results = at['targeter'].find_targets(data_graph)
                    for r in results:
                        found_node_targets.add(r)
                else:
                    results = at['qt'].find_targets(data_graph)
                    if not results or len(results.bindings) < 1:
                        continue
                    for r in results:
                        t = r['this']
                        found_node_targets.add(t)
        if debug:
            t2 = perf_counter()
            elapsed = t2 - t1
            self.logger.debug(f"Milliseconds to find focus nodes: {elapsed * 1000.0:.3f}ms")
        return found_node_targets

    @classmethod
    def make_focus_nodes_sparql_values(
        cls, target_classes_s: Set, implicit_classes_s: Set, target_objects_of_s: Set, target_subjects_of_s: Set
    ):
        init_bindings = {}
        values_keys = []
        values_vals = []
        if len(target_classes_s) > 1:
            values_keys.append("$targetClass")
            values_vals.append(list(target_classes_s))
        else:
            init_bindings["targetClass"] = next(iter(target_classes_s)) if len(target_classes_s) > 0 else "UNDEF"
        if len(implicit_classes_s) > 1:
            values_keys.append("$implicitClass")
            values_vals.append(list(implicit_classes_s))
        else:
            init_bindings["implicitClass"] = next(iter(implicit_classes_s)) if len(implicit_classes_s) > 0 else "UNDEF"
        if len(target_subjects_of_s) > 1:
            values_keys.append("$targetSubjectsOf")
            values_vals.append(list(target_subjects_of_s))
        else:
            init_bindings["targetSubjectsOf"] = (
                next(iter(target_subjects_of_s)) if len(target_subjects_of_s) > 0 else "UNDEF"
            )
        if len(target_objects_of_s) > 1:
            values_keys.append("$targetObjectsOf")
            values_vals.append(list(target_objects_of_s))
        else:
            init_bindings["targetObjectsOf"] = (
                next(iter(target_objects_of_s)) if len(target_objects_of_s) > 0 else "UNDEF"
            )
        if len(values_keys) < 1:
            return "", init_bindings
        else:
            values_clause = f"VALUES ({' '.join(values_keys)}) {{\n"
            product = itertools.product(*values_vals)
            for p in product:
                values_clause += f"\t( {' '.join(p_x.n3() for p_x in p)} )\n"
            values_clause += "}"
            return values_clause, init_bindings

    def focus_nodes_sparql(self, data_graph, debug=False):
        """
        The set of focus nodes for a shape may be identified as follows:

        specified in a shape using target declarations
        specified in any constraint that references a shape in parameters of shape-expecting constraint parameters (e.g. sh:node)
        specified as explicit input to the SHACL processor for validating a specific RDF term against a shape
        :return:
        """
        t1 = 0.0
        if debug:
            t1 = perf_counter()
        (target_nodes, target_classes, implicit_classes, target_objects_of, target_subjects_of) = self.target()
        if self._advanced:
            advanced_targets = self.advanced_target()
        else:
            advanced_targets = False
        found_node_targets = set()
        target_nodes = set(target_nodes)
        target_classes = set(target_classes)
        implicit_classes = set(implicit_classes)
        target_objects_of = set(target_objects_of)
        target_subjects_of = set(target_subjects_of)
        if all(
            (
                advanced_targets is False,
                len(target_nodes) < 1,
                len(target_classes) < 1,
                len(implicit_classes) < 1,
                len(target_objects_of) < 1,
                len(target_subjects_of) < 1,
            )
        ):
            return found_node_targets

        found_node_targets.update(target_nodes)
        if (
            advanced_targets is False
            and len(target_classes) < 1
            and len(implicit_classes) < 1
            and len(target_objects_of) < 1
            and len(target_subjects_of) < 1
        ):
            return found_node_targets
        if (
            len(target_classes) > 0
            or len(implicit_classes) > 0
            or len(target_objects_of) > 0
            or len(target_subjects_of) > 0
        ):
            focus_query = """\
            SELECT ?targetClass_F ?targetSubjectsOf_F ?targetObjectsOf_F WHERE {
                {VALUES_CLAUSE}
                OPTIONAL { {
                ?targetClass_F rdf:type/rdfs:subClassOf* $targetClass .
                } UNION {
                ?targetClass_F rdf:type/rdfs:subClassOf* $implicitClass .
                }. }
                OPTIONAL { ?targetSubjectsOf_F $targetSubjectsOf ?anyA . }
                OPTIONAL {?anyB $targetObjectsOf ?targetObjectsOf_F . }
            }
            """
            values_clause, init_bindings = self.make_focus_nodes_sparql_values(
                target_classes, implicit_classes, target_objects_of, target_subjects_of
            )
            new_query = focus_query.replace("{VALUES_CLAUSE}", values_clause)
            try:
                resp = data_graph.query(new_query, initBindings=init_bindings)
            except Exception as e:
                print(new_query)
                raise e
            if len(resp) > 0:
                for result_set in resp:
                    target_class_f, target_subjects_of_f, target_objects_of_f = result_set
                    if target_class_f is not None and target_class_f != "UNDEF":
                        found_node_targets.add(target_class_f)
                    if target_subjects_of_f is not None and target_subjects_of_f != "UNDEF":
                        found_node_targets.add(target_subjects_of_f)
                    if target_objects_of_f is not None and target_objects_of_f != "UNDEF":
                        found_node_targets.add(target_objects_of_f)
        if advanced_targets:
            for at_node, at in advanced_targets.items():
                if at['type'] == SH_SPARQLTarget:
                    qh = at['qh']
                    select = qh.apply_prefixes(qh.select_text)
                    results = data_graph.query(select, initBindings=None)
                    if not results or len(results.bindings) < 1:
                        continue
                    for r in results:
                        t = r['this']
                        found_node_targets.add(t)
                elif at['type'] in (SH_JSTarget, SH_JSTargetType):
                    raise ReportableRuntimeError(
                        "SHACL Advanced Targets with JSTargets are not yet implemented in SPARQL Remote Graph Mode."
                    )
                else:
                    results = at['qt'].find_targets(data_graph)
                    if not results or len(results.bindings) < 1:
                        continue
                    for r in results:
                        t = r['this']
                        found_node_targets.add(t)
        if debug:
            t2 = perf_counter()
            elapsed = t2 - t1
            self.logger.debug(f"Milliseconds to find focus nodes: {elapsed * 1000.0:.3f}ms")
        return found_node_targets

    def value_nodes(self, target_graph, focus, sparql_mode: bool = False, debug: bool = False):
        """
        For each focus node, you can get a set of value nodes.
        For a Node Shape, each focus node has just one value node,
            which is just the focus_node
        :param target_graph:
        :param focus:
        :param sparql_mode:
        :type sparql_mode: bool
        :param debug:
        :type debug: bool
        :return:
        """
        t1 = 0.0
        if debug:
            t1 = perf_counter()
        if not isinstance(focus, (tuple, list, set)):
            focus = [focus]
        if not self.is_property_shape:
            if debug:
                t2 = perf_counter()
                elapsed = t2 - t1
                self.logger.debug(f"Milliseconds to find value nodes for focus nodes: {elapsed * 1000.0:.3f}ms")
            return {f: set((f,)) for f in focus}
        path_val = self.path()

        focus_dict: Dict[RDFNode, Set[RDFNode]] = {}
        if sparql_mode:
            # Shortcut for simple URI path, path rewriting and everything else
            if isinstance(path_val, URIRef):
                sparql_path = path_val.n3(namespace_manager=target_graph.namespace_manager)
            else:
                prefixes = dict(target_graph.namespace_manager.namespaces())
                sparql_path = shacl_path_to_sparql_path(self.sg, path_val, prefixes=prefixes)
            values_query = f"SELECT {' '.join(f'?v{i}' for i, _ in enumerate(focus))} WHERE {{\n"
            init_bindings = {}
            for i, f in enumerate(focus):
                focus_dict[f] = set()
                values_query += f"OPTIONAL {{ \t$f{i} {sparql_path} ?v{i} . }}\n"
                init_bindings[f"f{i}"] = f
            values_query += "}"
            try:
                results = target_graph.query(values_query, initBindings=init_bindings)
            except Exception as e:
                print(e)
                raise
            if len(results) > 0:
                for r in results:
                    for i, f in enumerate(focus):
                        row_focus_result = r[i]
                        if row_focus_result is None or row_focus_result == "UNDEF":
                            continue
                        focus_dict[f].add(row_focus_result)
            else:
                pass
        else:
            for f in focus:
                focus_dict[f] = value_nodes_from_path(self.sg, f, path_val, target_graph)
        if debug:
            t2 = perf_counter()
            elapsed = t2 - t1
            self.logger.debug(f"Milliseconds to find value nodes for focus nodes: {elapsed * 1000.0:.3f}ms")
        return focus_dict

    def find_custom_constraints(self):
        applicable_custom_constraints = set()
        for c in self.sg.custom_constraints:
            mandatory = (p for p in c.parameters if not p.optional)
            found_all_mandatory = True
            for mandatory_param in mandatory:
                path = mandatory_param.path()
                assert isinstance(path, URIRef)
                found_vals = set(self.sg.objects(self.node, path))
                # found_vals = value_nodes_from_path(self.node, mandatory_param.path(), self.sg.graph)
                found_all_mandatory = found_all_mandatory and bool(len(found_vals) > 0)
            if found_all_mandatory:
                applicable_custom_constraints.add(c)
        return applicable_custom_constraints

    def validate(
        self,
        executor: SHACLExecutor,
        target_graph: GraphLike,
        focus: Optional[
            Union[
                Sequence[RDFNode],
                RDFNode,
            ]
        ] = None,
        _evaluation_path: Optional[List] = None,
    ):
        if self.deactivated:
            if executor.debug:
                self.logger.debug(f"Skipping shape because it is deactivated: {str(self)}")
            return True, []
        focus_list: Sequence[RDFNode]
        if focus is not None:
            lh_shape = False
            rh_shape = True
            self.logger.debug(f"Running evaluation of Shape {str(self)}")
            # Passed in Focus node _can_ be a Literal, happens in PropertyShapes
            # when the path resolves to a literal or set of Literals
            if isinstance(focus, (IdentifiedNode, Literal)):
                focus_list = [focus]
            else:
                focus_list = list(focus)
            self.logger.debug(f"Shape was passed {len(focus_list)} Focus Node/s to evaluate.")
        else:
            lh_shape = True
            rh_shape = False
            self.logger.debug(f"Checking if Shape {str(self)} defines its own targets.")
            self.logger.debug("Identifying targets to find focus nodes.")
            if executor.sparql_mode:
                focus_set = self.focus_nodes_sparql(target_graph, debug=executor.debug)
            else:
                focus_set = self.focus_nodes(target_graph, debug=executor.debug)
            focus_list = list(focus_set)
            self.logger.debug(f"Found {len(focus_list)} Focus Nodes to evaluate.")

        if len(focus_list) < 1:
            # It's possible for shapes to have _no_ focus nodes
            # (they are called in other ways)
            if executor.debug:
                self.logger.debug(f"Skipping shape {str(self)} because it found no focus nodes.")
            return True, []
        else:
            self.logger.debug(f"Running evaluation of Shape {str(self)}")

        if executor.focus_nodes is not None and len(executor.focus_nodes) > 0:
            filtered_focus_nodes: List[Union[URIRef]] = []
            for _fo in focus_list:  # type: RDFNode
                if isinstance(_fo, URIRef) and _fo in executor.focus_nodes:
                    filtered_focus_nodes.append(_fo)
            len_orig_focus = len(focus_list)
            len_filtered_focus = len(filtered_focus_nodes)
            if len_filtered_focus < 1:
                self.logger.debug(f"Skipping shape {str(self)} because specified focus nodes are not targeted.")
                return True, []
            elif len_filtered_focus != len_orig_focus:
                self.logger.debug(
                    f"Filtered focus nodes based on focus_nodes option. Only {len_filtered_focus} of {len_orig_focus} focus nodes remain."
                )
            focus_list = filtered_focus_nodes
        t1 = ct1 = 0.0  # prevent warnings about use-before-assign
        collect_stats = bool(executor.debug)

        if _evaluation_path is None:
            _evaluation_path = []
        else:
            validation_depth = len(_evaluation_path) // 2

            if validation_depth >= executor.max_validation_depth:
                # depth of 14 (_evaluation_length=28) is the depth required to
                # successfully do the meta-shacl test on shacl.ttl
                path_str = " -> ".join((str(e) for e in _evaluation_path))
                raise ReportableRuntimeError("Validation path too deep!\n{}".format(path_str))
        if collect_stats:
            t1 = perf_counter()
        # Lazy import here to avoid an import loop
        CONSTRAINT_PARAMETERS, PARAMETER_MAP = getattr(module, 'CONSTRAINT_PARAMS', (None, None))
        if not CONSTRAINT_PARAMETERS or not PARAMETER_MAP:
            from .constraints import ALL_CONSTRAINT_PARAMETERS, CONSTRAINT_PARAMETERS_MAP

            setattr(module, 'CONSTRAINT_PARAMS', (ALL_CONSTRAINT_PARAMETERS, CONSTRAINT_PARAMETERS_MAP))
            CONSTRAINT_PARAMETERS = ALL_CONSTRAINT_PARAMETERS
            PARAMETER_MAP = CONSTRAINT_PARAMETERS_MAP
        if self.sg.js_enabled or self._advanced:
            search_parameters = CONSTRAINT_PARAMETERS.copy()
            constraint_map = PARAMETER_MAP.copy()
            if self._advanced:
                from pyshacl.constraints.advanced import ExpressionConstraint, SH_expression

                search_parameters.append(SH_expression)
                constraint_map[SH_expression] = ExpressionConstraint
            if self.sg.js_enabled:
                from pyshacl.extras.js.constraint import JSConstraint, SH_js

                search_parameters.append(SH_js)
                constraint_map[SH_js] = JSConstraint
        else:
            search_parameters = CONSTRAINT_PARAMETERS
            constraint_map = PARAMETER_MAP
        parameters = (p for p, v in self.sg.predicate_objects(self.node) if p in search_parameters)
        reports = []
        focus_value_nodes = self.value_nodes(
            target_graph, focus_list, sparql_mode=executor.sparql_mode, debug=executor.debug
        )
        filter_reports: bool = False
        allow_conform: bool = False
        allowed_severities: Set[URIRef] = set()
        if executor.allow_infos:
            allowed_severities.add(SH_Info)
        if executor.allow_warnings:
            allowed_severities.add(SH_Info)
            allowed_severities.add(SH_Warning)
        if executor.allow_infos or executor.allow_warnings:
            if self.severity in allowed_severities:
                allow_conform = True
            else:
                filter_reports = True

        non_conformant = False
        done_constraints = set()
        run_count = 0
        _evaluation_path.append(self)
        if executor.debug:
            path_str = " -> ".join((str(e) for e in _evaluation_path))
            self.logger.debug(f"Current shape evaluation path: {path_str}")
        constraint_components = [constraint_map[p] for p in iter(parameters)]
        constraint_component: Type['ConstraintComponent']
        for constraint_component in constraint_components:
            if constraint_component in done_constraints:
                continue
            try:
                # if executor.debug:
                #     self.logger.debug(f"Constructing Constraint Component: {repr(constraint_component)}")
                c = constraint_component(self)
            except ConstraintLoadWarning as w:
                self.logger.warning(repr(w))
                continue
            except ConstraintLoadError as e:
                self.logger.error(repr(e))
                raise e
            _e_p_copy = _evaluation_path[:]
            _e_p_copy.append(c)
            if executor.debug:
                self.logger.debug(f"Checking conformance for constraint: {str(c)}")
            if collect_stats:
                ct1 = perf_counter()
            if executor.debug:
                path_str = " -> ".join((str(e) for e in _e_p_copy))
                self.logger.debug(f"Current constraint evaluation path: {path_str}")
            _is_conform, _reports = c.evaluate(executor, target_graph, focus_value_nodes, _e_p_copy)
            if executor.debug:
                if collect_stats:
                    ct2 = perf_counter()
                    elapsed = ct2 - ct1
                    self.logger.debug(f"Milliseconds to check constraint {str(c)}: {elapsed * 1000.0:.3f}ms")
                if _is_conform:
                    self.logger.debug(f"DataGraph conforms to constraint {c}.")
                elif allow_conform:
                    self.logger.debug(f"Focus nodes do _not_ conform to constraint {c} but given severity is allowed.")
                else:
                    self.logger.debug(f"Focus nodes do _not_ conform to constraint {c}.")
                    if lh_shape or (not rh_shape):
                        for v_str, v_node, v_parts in _reports:
                            self.logger.debug(v_str)

            if _is_conform or allow_conform:
                ...
            elif filter_reports:
                all_allow = True
                for v_str, v_node, v_parts in _reports:
                    severity_bits = list(filter(lambda p: p[0] == v_node and p[1] == SH_resultSeverity, v_parts))
                    if severity_bits:
                        all_allow = all_allow and (severity_bits[0][2] in allowed_severities)
                non_conformant = non_conformant or (not all_allow)
            else:
                non_conformant = non_conformant or (not _is_conform)
            reports.extend(_reports)
            run_count += 1
            done_constraints.add(constraint_component)
            if non_conformant and executor.abort_on_first:
                break
        applicable_custom_constraints = self.find_custom_constraints()
        for a in applicable_custom_constraints:
            if non_conformant and executor.abort_on_first:
                break
            _e_p_copy2 = _evaluation_path[:]
            validator = a.make_validator_for_shape(self)
            _e_p_copy2.append(validator)
            _is_conform, _r = validator.evaluate(executor, target_graph, focus_value_nodes, _e_p_copy2)
            non_conformant = non_conformant or (not _is_conform)
            reports.extend(_r)
            run_count += 1
        if collect_stats:
            t2 = perf_counter()
            elapsed = t2 - t1
            self.logger.debug(f"Milliseconds to evaluate shape {str(self)}: {elapsed * 1000.0:.3f}ms")
        # print(_evaluation_path, "Passes" if not non_conformant else "Fails")
        return (not non_conformant), reports



# -*- coding: utf-8 -*-
#
import logging
from os import getenv
from typing import Any, Dict, List, Optional, Sequence, Union

import rdflib
from rdflib import URIRef

from .consts import (
    env_truths,
)
from .extras import check_extra_installed
from .functions import apply_functions, gather_functions, unapply_functions
from .pytypes import GraphLike, SHACLExecutor
from .rdfutil import (
    clone_graph,
    inoculate,
    inoculate_dataset,
    mix_datasets,
    mix_graphs,
)
from .rules import apply_rules, gather_rules
from .run_type import PySHACLRunType
from .shapes_graph import ShapesGraph
from .target import apply_target_types, gather_target_types

USE_FULL_MIXIN = getenv("PYSHACL_USE_FULL_MIXIN") in env_truths


class RuleExpandRunner(PySHACLRunType):
    def __init__(
        self,
        data_graph: GraphLike,
        *args,
        shacl_graph: Optional[GraphLike] = None,
        ont_graph: Optional[GraphLike] = None,
        options: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        options = options or {}
        self._load_default_options(options)
        self.options = options  # type: dict
        self.logger = options['logger']  # type: logging.Logger
        self.debug = options['debug']
        self.pre_inferenced = kwargs.pop('pre_inferenced', False)
        self.inplace = options['inplace']
        if not isinstance(data_graph, rdflib.Graph):
            raise RuntimeError("data_graph must be a rdflib Graph-like object")
        self.data_graph = data_graph  # type: GraphLike
        self._target_graph: Union[GraphLike, None] = None
        self.ont_graph = ont_graph  # type: Optional[GraphLike]
        self.data_graph_is_multigraph = isinstance(self.data_graph, (rdflib.Dataset, rdflib.ConjunctiveGraph))
        if self.ont_graph is not None and isinstance(self.ont_graph, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
            self.ont_graph.default_union = True
        if shacl_graph is None:
            shacl_graph = clone_graph(data_graph, identifier='shacl')
        assert isinstance(shacl_graph, rdflib.Graph), "shacl_graph must be a rdflib Graph object"
        self.shacl_graph = ShapesGraph(shacl_graph, self.debug, self.logger)  # type: ShapesGraph

        if options['use_js']:
            is_js_installed = check_extra_installed('js')
            if is_js_installed:
                self.shacl_graph.enable_js()

    @classmethod
    def _load_default_options(cls, options_dict: dict):
        options_dict.setdefault('debug', False)
        options_dict.setdefault('inference', 'none')
        options_dict.setdefault('inplace', False)
        options_dict.setdefault('use_js', False)
        options_dict.setdefault('iterate_rules', False)
        options_dict.setdefault('focus_nodes', None)
        options_dict.setdefault('use_shapes', None)
        if 'logger' not in options_dict:
            options_dict['logger'] = logging.getLogger(__name__)
            if options_dict['debug']:
                options_dict['logger'].setLevel(logging.DEBUG)

    @property
    def target_graph(self) -> Union[GraphLike, None]:
        return self._target_graph

    def mix_in_ontology(self):
        if USE_FULL_MIXIN:
            if not self.data_graph_is_multigraph:
                return mix_graphs(self.data_graph, self.ont_graph, "inplace" if self.inplace else None)
            return mix_datasets(self.data_graph, self.ont_graph, "inplace" if self.inplace else None)
        if not self.data_graph_is_multigraph:
            if self.inplace:
                to_graph = self.data_graph
            else:
                to_graph = clone_graph(self.data_graph, identifier=self.data_graph.identifier)
            return inoculate(to_graph, self.ont_graph)
        return inoculate_dataset(
            self.data_graph,
            self.ont_graph,
            self.data_graph if self.inplace else None,
            URIRef("urn:pyshacl:inoculation"),
        )

    def make_executor(self) -> SHACLExecutor:
        return SHACLExecutor(
            validator=self,
            advanced_mode=True,
            abort_on_first=False,
            allow_infos=False,
            allow_warnings=False,
            iterate_rules=bool(self.options.get("iterate_rules", False)),
            sparql_mode=False,
            max_validation_depth=999,
            focus_nodes=self.options.get("focus_nodes", None),
            debug=self.debug,
        )

    def run(self) -> GraphLike:
        datagraph: Union[GraphLike, None] = self.target_graph
        if datagraph is not None:
            # Target graph is already set up with pre-inferenced and pre-cloned data_graph
            self._target_graph = datagraph
        else:
            has_cloned = False
            if self.ont_graph is not None:
                if self.inplace:
                    self.logger.debug("Adding ontology definitions to DataGraph")
                else:
                    self.logger.debug("Cloning DataGraph to temporary memory graph, to add ontology definitions.")
                # creates a copy of self.data_graph, doesn't modify it
                datagraph = self.mix_in_ontology()
                has_cloned = True
            else:
                datagraph = self.data_graph
            inference_option = self.options.get('inference', 'none')
            if self.inplace and self.debug:
                self.logger.debug("Skipping DataGraph clone because PySHACL is operating in inplace mode.")
            if inference_option and not self.pre_inferenced and str(inference_option) != "none":
                if not has_cloned and not self.inplace:
                    self.logger.debug("Cloning DataGraph to temporary memory graph before pre-inferencing.")
                    datagraph = clone_graph(datagraph)
                    has_cloned = True
                self.logger.debug(f"Running pre-inferencing with option='{inference_option}'.")
                self._run_pre_inference(
                    datagraph, inference_option, URIRef("urn:pyshacl:inference"), logger=self.logger
                )
                self.pre_inferenced = True
            if not has_cloned and not self.inplace:
                # We still need to clone in advanced mode, because of triple rules
                self.logger.debug(
                    "Forcing clone of DataGraph because expanding rules cannot modify the input datagraph."
                )
                datagraph = clone_graph(datagraph)
                has_cloned = True
            self._target_graph = datagraph
        assert self._target_graph is not None
        if self.options.get("use_shapes", None) is not None and len(self.options["use_shapes"]) > 0:
            using_manually_specified_shapes = True
            expanded_use_shapes = []
            for s in self.options["use_shapes"]:
                s_lower = s.lower()
                if (
                    s_lower.startswith("http:")
                    or s_lower.startswith("https:")
                    or s_lower.startswith("urn:")
                    or s_lower.startswith("file:")
                ):
                    expanded_use_shapes.append(URIRef(s))
                else:
                    try:
                        expanded_use_shape = self.shacl_graph.graph.namespace_manager.expand_curie(s)
                    except ValueError:
                        expanded_use_shape = URIRef(s)
                    expanded_use_shapes.append(expanded_use_shape)
            shapes = self.shacl_graph.shapes_from_uris(expanded_use_shapes)
        else:
            using_manually_specified_shapes = False
            shapes = self.shacl_graph.shapes  # This property getter triggers shapes harvest.
        option_focus_nodes = self.options.get("focus_nodes", None)
        if option_focus_nodes is not None and len(option_focus_nodes) > 0:
            # Expand any CURIEs in the focus_nodes list
            expanded_focus_nodes: List[URIRef] = []
            for f in option_focus_nodes:
                f_lower = f.lower()
                if (
                    f_lower.startswith("http:")
                    or f_lower.startswith("https:")
                    or f_lower.startswith("urn:")
                    or f_lower.startswith("file:")
                ):
                    expanded_focus_nodes.append(URIRef(f))
                else:
                    try:
                        expanded_focus_node = self._target_graph.namespace_manager.expand_curie(f)
                    except ValueError:
                        expanded_focus_node = URIRef(f)
                    expanded_focus_nodes.append(expanded_focus_node)
            self.options["focus_nodes"] = expanded_focus_nodes
            specified_focus_nodes: Union[None, Sequence[URIRef]] = expanded_focus_nodes
        else:
            specified_focus_nodes = None
        executor = self.make_executor()
        # Special hack, if we are using manually specified shapes, and have
        # manually specified focus nodes, then we need to disable the
        # focus_nodes in the executor, because we apply the specified focus
        # nodes directly to the specified shapes.
        if using_manually_specified_shapes and specified_focus_nodes is not None:
            executor.focus_nodes = None
        self.logger.debug("Activating SHACL-AF Features.")
        target_types = gather_target_types(self.shacl_graph)
        gather_from_shapes = None if not using_manually_specified_shapes else [s.node for s in shapes]
        gathered_functions = gather_functions(executor, self.shacl_graph)
        gathered_rules = gather_rules(executor, self.shacl_graph, from_shapes=gather_from_shapes)

        for s in shapes:
            s.set_advanced(True)
        apply_target_types(target_types)
        if isinstance(self._target_graph, (rdflib.Dataset, rdflib.ConjunctiveGraph)):
            self._target_graph.default_union = True

        g = self._target_graph

        if specified_focus_nodes is not None and using_manually_specified_shapes:
            on_focus_nodes: Union[Sequence[URIRef], None] = specified_focus_nodes
        else:
            on_focus_nodes = None

        if self.debug:
            self.logger.debug(f"Running SHACL Rules on DataGraph named {g.identifier}")
        if gathered_functions:
            apply_functions(executor, gathered_functions, g)
        try:
            if gathered_rules:
                apply_rules(executor, gathered_rules, g, focus_nodes=on_focus_nodes)
        finally:
            if gathered_functions:
                unapply_functions(gathered_functions, g)

        return g



# -*- coding: utf-8 -*-
#
import os
import platform
import subprocess
import sys
from os import getenv, path
from sys import stderr

from rdflib import RDF, Graph, URIRef

PATH = getenv("PATH", "")
PP = getenv('PYTHONPATH', "")
here_dir = path.abspath(path.dirname(__file__))
ENV_VARS = {"PATH": PATH, "PYTHONPATH": ':'.join((here_dir, PP))}
PH = getenv('PYTHONHOME', "")
if PH:
    ENV_VARS['PYTHONHOME'] = PH
VE = getenv('VIRTUAL_ENV', "")
if VE:
    ENV_VARS['VIRTUAL_ENV'] = VE
    virtual_bin = path.join(VE, "bin")
    ENV_VARS['PATH'] = ':'.join((virtual_bin, PATH))
abs_resources_dir = path.join(here_dir, 'resources')
cmdline_files_dir = path.join(abs_resources_dir, 'cmdline_tests')

check_resources = path.join(path.abspath(os.getcwd()), 'resources')
in_test_dir = False
if path.exists(check_resources) and path.isdir(check_resources):
    in_test_dir = True
else:
    in_test_dir = False

if in_test_dir:
    lib_dir = os.path.abspath(os.path.join(here_dir, os.pardir))
    ENV_VARS["PYTHONPATH"] = ':'.join((lib_dir, PP))

it = ENV_VARS["PYTHONPATH"].split(":")
scr_dir = "scripts-{}.{}".format(sys.version_info[0], sys.version_info[1])
if in_test_dir:
    scr_dir = path.join('..', scr_dir)
check_scrdir = path.join(path.abspath(os.getcwd()), scr_dir)
if path.exists(check_scrdir) and path.isdir(check_scrdir):
    has_scripts_dir = True
else:
    has_scripts_dir = False

bin_dir = "bin"
if in_test_dir:
    bin_dir = path.join('..', bin_dir)
check_bindir = path.join(path.abspath(os.getcwd()), bin_dir)
if path.exists(check_bindir) and path.isdir(check_bindir):
    has_bin_dir = True
else:
    has_bin_dir = False

cli_rules_script = "pyshacl/cli_rules.py"
if in_test_dir:
    cli_rules_script = path.join('..', cli_rules_script)
check_cli_script = path.join(path.abspath(os.getcwd()), cli_rules_script)
if path.exists(check_cli_script) and path.isfile(check_cli_script):
    has_cli_script = True
else:
    has_cli_script = False

if has_scripts_dir:
    pyshacl_rules_command = ["{}/pyshacl_rules".format(scr_dir)]
elif has_bin_dir:
    pyshacl_rules_command = ["{}/pyshacl_rules".format(bin_dir)]
elif has_cli_script:
    pyshacl_rules_command = ["python3", cli_rules_script]
else:
    pyshacl_rules_command = ["pyshacl_rules"]


def test_cmdline_rules():
    if not hasattr(subprocess, 'run'):
        print("Subprocess.run() not available, skip this test")
        assert True
        return True
    if platform.system() == "Windows":
        print("Commandline tests cannot run on Windows.")
        assert True
        return True
    if os.environ.get("PYBUILD_NAME", None) is not None:
        print("We don't have access to scripts dir during pybuild process.")
        assert True
        return True
    graph_file = path.join(cmdline_files_dir, 'rules_d.ttl')
    shacl_file = path.join(cmdline_files_dir, 'rules_s.ttl')
    cmd = pyshacl_rules_command
    args = [graph_file, '-s', shacl_file, '-i', 'rdfs']
    res = subprocess.run(cmd + args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=ENV_VARS)
    print("result = {}".format(res.returncode))
    output_unicode = res.stdout.decode('utf-8')
    print(res.stderr.decode('utf-8'))
    assert res.returncode == 0
    output_g = Graph().parse(data=output_unicode, format='trig')
    person_classes = set(
        output_g.objects(
            URIRef("http://datashapes.org/shasf/tests/expression/rules.test.data#Jenny"), predicate=RDF.type
        )
    )
    assert URIRef("http://datashapes.org/shasf/tests/expression/rules.test.ont#Administrator") in person_classes
    assert URIRef("http://datashapes.org/shasf/tests/expression/rules.test.ont#Person") in person_classes


if __name__ == "__main__":
    test_cmdline_rules()
